# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ProviderError(Exception):
    pass


class StateError(Exception):
    pass


def validate_state_empty(state):
    """
    validate_empty_state ensures that the provided state file
    is empty
    """

    try:
        if len(state["resources"]) > 0:
            return False
        else:
            return True
    except KeyError:
        raise StateError("resources key does not exist in state!")


def validate_state_region(state):
    """
    validate_state_region validates that a statefile
    was previously used in the region the current
    deployment is being created for
    """



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import shlex
import shutil
import subprocess
import tempfile
from collections import OrderedDict

import boto3
import click
import yaml


class State(object):
    """A module to hold the state of the application."""

    def __init__(self, args=None, config_file=None, clean=True):
        """Setup state with args that are passed."""
        self.clean = clean
        self.temp_dir = tempfile.mkdtemp()
        self.args = self.StateArgs()

        if args is not None:
            self.add_args(args)

        if config_file is not None:
            self.load_config(config_file)

    def __del__(self):
        """Cleanup the temporary directory after execution."""
        if self.clean:
            shutil.rmtree(self.temp_dir)

    def add_args(self, args):
        """Add a dictionary of args."""
        for k, v in args.items():
            self.add_arg(k, v)

    def add_arg(self, k, v):
        """Add an argument to the state args."""
        setattr(self.args, k, v)
        return None

    def load_config(self, config_file):
        try:
            with open(config_file, "r") as cfile:
                self.config = ordered_config_load(cfile, self.args)
        except IOError:
            click.secho(
                "Unable to open configuration file: {}".format(config_file), fg="red"
            )
            raise SystemExit(1)

    class StateArgs(object):
        """A class to hold arguments in the state for easier access."""

        pass


def create_table(
    name, region, key_id, key_secret, session_token, read_capacity=1, write_capacity=1
):
    """Create a dynamodb table."""
    client = boto3.client(
        "dynamodb",
        region_name=region,
        aws_access_key_id=key_id,
        aws_secret_access_key=key_secret,
        aws_session_token=session_token,
    )
    tables = client.list_tables()
    table_key = "LockID"
    if name in tables["TableNames"]:
        click.secho("DynamoDB lock table found, continuing.", fg="yellow")
    else:
        click.secho(
            "DynamoDB lock table not found, creating, please wait...", fg="yellow"
        )
        client.create_table(
            TableName=name,
            KeySchema=[{"AttributeName": table_key, "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": table_key, "AttributeType": "S"},],
            ProvisionedThroughput={
                "ReadCapacityUnits": read_capacity,
                "WriteCapacityUnits": write_capacity,
            },
        )

        client.get_waiter("table_exists").wait(
            TableName=name, WaiterConfig={"Delay": 10, "MaxAttempts": 30}
        )


def get_aws_id(key_id, key_secret, session_token=None):
    """Return the AWS account ID."""
    client = boto3.client(
        "sts",
        aws_access_key_id=key_id,
        aws_secret_access_key=key_secret,
        aws_session_token=session_token,
    )
    return client.get_caller_identity()["Account"]


def ordered_config_load(
    stream, args, Loader=yaml.SafeLoader, object_pairs_hook=OrderedDict
):
    """
    Load a yaml config, and replace templated items.

    Derived from:
    https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
    """

    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        for item in node.value:
            if isinstance(item, tuple):
                for oneitem in item:
                    if isinstance(oneitem, yaml.ScalarNode):
                        oneitem.value = replace_vars(oneitem.value, args)

        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping
    )
    return yaml.load(stream, OrderedLoader)


def replace_vars(var, args):
    """Replace variables with template values."""
    var_pattern = r"//\s*(\S*)\s*//"
    match = re.match(var_pattern, var, flags=0)
    if not match:
        return var
    try:
        var = getattr(args, match.group(1).replace("-", "_"))
    except AttributeError:
        raise (ValueError("substitution not found for {}".format(var)))
    return var


def pipe_exec(args, stdin=None, cwd=None, env=None):
    """
    A function to accept a list of commands and pipe them together.

    Takes optional stdin to give to the first item in the pipe chain.
    """
    count = 0
    commands = []
    if env is None:
        env = os.environ.copy()

    if not isinstance(args, list):
        args = [args]

    for i in args:
        if count == 0:
            if stdin is None:
                commands.append(
                    subprocess.Popen(
                        shlex.split(i),
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        cwd=cwd,
                        env=env,
                    )
                )
            else:
                commands.append(
                    subprocess.Popen(
                        shlex.split(i),
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        stdin=subprocess.PIPE,
                        cwd=cwd,
                        env=env,
                    )
                )
        else:
            commands.append(
                subprocess.Popen(
                    shlex.split(i),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    stdin=commands[count - 1].stdout,
                    cwd=cwd,
                    env=env,
                )
            )
        count = count + 1

    if stdin is not None:
        stdin_bytes = stdin.encode()
        if len(commands) > 1:
            commands[0].communicate(input=stdin_bytes)
            stdout, stderr = commands[-1].communicate()
            commands[-1].wait()
            returncode = commands[-1].returncode
        else:
            stdout, stderr = commands[0].communicate(input=stdin_bytes)
            commands[0].wait()
            returncode = commands[0].returncode
    else:
        stdout, stderr = commands[-1].communicate()
        commands[-1].wait()
        returncode = commands[-1].returncode

    return (returncode, stdout, stderr)



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import copy
import glob
import json
import os
import re
import shutil
import sys
import urllib
import zipfile

from pathlib import Path

import click
import jinja2

from tfworker.main import pipe_exec


class TerraformError(Exception):
    pass


class PlanChange(Exception):
    pass


class HookError(Exception):
    pass


def download_plugins(plugins, temp_dir):
    """
    Download the required plugins.

    This could be further optimized to not download plugins from hashicorp, but rather have them in a local repository
    or host them in s3, and get them from an internal s3 endpoint so no transit charges are incurred. Ideally these
    would be stored between runs, and only downloaded if the versions have changed. In production try  to remove all
    all external repositories/sources from the critical path.
    """
    platform = "{}_{}".format(sys.platform.rstrip("2"), "amd64")
    plugin_dir = "{}/terraform-plugins".format(temp_dir)
    if not os.path.isdir(plugin_dir):
        os.mkdir(plugin_dir)

    for name, details in plugins.items():
        # Get platform and strip 2 off linux 2.x kernels
        file_name = "terraform-provider-{}_{}_{}.zip".format(
            name, details["version"], platform
        )
        uri = "https://releases.hashicorp.com/terraform-provider-{}/{}/{}".format(
            name, details["version"], file_name
        )
        click.secho(
            "getting plugin: {} version {} from {}".format(
                name, details["version"], uri
            ),
            fg="yellow",
        )

        with urllib.request.urlopen(uri) as response, open(
            "{}/{}".format(plugin_dir, file_name), "wb"
        ) as plug_file:
            shutil.copyfileobj(response, plug_file)
        with zipfile.ZipFile("{}/{}".format(plugin_dir, file_name)) as zip_file:
            zip_file.extractall(plugin_dir)
        os.remove("{}/{}".format(plugin_dir, file_name))

        files = glob.glob("{}/terraform-provider*".format(plugin_dir))
        for afile in files:
            os.chmod(afile, 0o755)


def prep_modules(src, dst):
    """Puts the modules sub directories into place."""
    mod_source = "{}/terraform-modules".format(src).replace("//", "/")
    mod_destination = "{}/terraform-modules".format(dst).replace("//", "/")
    shutil.copytree(
        mod_source,
        mod_destination,
        symlinks=True,
        ignore=shutil.ignore_patterns("test", ".terraform", "terraform.tfstate*"),
    )


def prep_def(name, definition, all_defs, temp_dir, repo_path, deployment, args):
    """ prepare the definitions for running """
    repo = Path("{}/{}".format(repo_path, definition["path"]).replace("//", "/"))
    target = Path("{}/definitions/{}".format(temp_dir, name).replace("//", "/"))
    target.mkdir(parents=True, exist_ok=True)

    if not repo.exists():
        click.secho(
            "Error preparing definition {}, path {} does not exist".format(
                name, repo.resolve()
            ),
            fg="red",
        )
        sys.exit(1)

    # Prepare variables
    terraform_vars = None
    template_vars = make_vars("template_vars", definition, all_defs)
    terraform_vars = make_vars("terraform_vars", definition, all_defs)
    locals_vars = make_vars("remote_vars", definition)
    template_vars["deployment"] = deployment
    terraform_vars["deployment"] = deployment

    # Put terraform files in place
    for tf in repo.glob("*.tf"):
        shutil.copy("{}".format(str(tf)), str(target))
    for tf in repo.glob("*.tfvars"):
        shutil.copy("{}".format(str(tf)), str(target))
    if os.path.isdir(str(repo) + "/templates".replace("//", "/")):
        shutil.copytree(
            "{}/templates".format(str(repo)), "{}/templates".format(str(target))
        )
    if os.path.isdir(str(repo) + "/policies".replace("//", "/")):
        shutil.copytree(
            "{}/policies".format(str(repo)), "{}/policies".format(str(target))
        )
    if os.path.isdir(str(repo) + "/scripts".replace("//", "/")):
        shutil.copytree(
            "{}/scripts".format(str(repo)), "{}/scripts".format(str(target))
        )
    if os.path.isdir(str(repo) + "/hooks".replace("//", "/")):
        shutil.copytree("{}/hooks".format(str(repo)), "{}/hooks".format(str(target)))
    if os.path.isdir(str(repo) + "/repos".replace("//", "/")):
        shutil.copytree("{}/repos".format(str(repo)), "{}/repos".format(str(target)))

    # Render jinja templates and put in place
    env = jinja2.Environment(loader=jinja2.FileSystemLoader)

    for j2 in repo.glob("*.j2"):
        contents = env.get_template(str(j2)).render(**template_vars)
        with open("{}/{}".format(str(target), str(j2)), "w+") as j2_file:
            j2_file.write(contents)

    # Create local vars from remote data sources
    if len(list(locals_vars.keys())) > 0:
        with open("{}/{}".format(str(target), "worker-locals.tf"), "w+") as tflocals:
            tflocals.write("locals {\n")
            for k, v in locals_vars.items():
                tflocals.write(
                    '  {} = "${{data.terraform_remote_state.{}}}"\n'.format(k, v)
                )
            tflocals.write("}\n\n")

    # Create the terraform configuration, terraform.tf
    state = render_remote_state(name, deployment, args)
    remote_data = render_remote_data_sources(all_defs["definitions"], name, args)
    providers = render_providers(all_defs["providers"], args)
    with open("{}/{}".format(str(target), "terraform.tf"), "w+") as tffile:
        tffile.write("{}\n\n".format(providers))
        tffile.write("{}\n\n".format(state))
        tffile.write("{}\n\n".format(remote_data))

    # Create the variable definitions
    with open("{}/{}".format(str(target), "worker.auto.tfvars"), "w+") as varfile:
        for k, v in terraform_vars.items():
            if isinstance(v, list):
                varstring = "[{}]".format(", ".join(map(quote_str, v)))
                varfile.write("{} = {}\n".format(k, varstring))
            else:
                varfile.write('{} = "{}"\n'.format(k, v))


def make_vars(section, single, base=None):
    """Make a variables dictionary based on default vars, as well as specific vars for an item."""
    if base is None:
        base = {}

    item_vars = copy.deepcopy(base.get(section, {}))
    for k, v in single.get(section, {}).items():
        # terraform expects variables in a specific type, so need to convert bools to a lower case true/false
        matched_type = False
        if v is True:
            item_vars[k] = "true"
            matched_type = True
        if v is False:
            item_vars[k] = "false"
            matched_type = True
        if not matched_type:
            item_vars[k] = v

    return item_vars


def quote_str(string):
    """Put literal quotes around a string."""
    return '"{}"'.format(string)


def render_remote_state(name, deployment, args):
    """Return remote for the definition."""
    state_config = []
    state_config.append("terraform {")
    state_config.append('  backend "s3" {')
    state_config.append('    region = "{}"'.format(args.state_region))
    state_config.append('    bucket = "{}"'.format(args.s3_bucket))
    state_config.append(
        '    key = "{}/{}/terraform.tfstate"'.format(args.s3_prefix, name)
    )
    state_config.append('    dynamodb_table = "terraform-{}"'.format(deployment))
    state_config.append('    encrypt = "true"')
    state_config.append("  }")
    state_config.append("}")
    return "\n".join(state_config)


def render_remote_data_sources(definitions, exclude, args):
    """Return remote data sources for all items until the excluded item"""
    remote_data_config = []
    for name, _ in definitions.items():
        if name == exclude:
            break
        remote_data_config.append('data "terraform_remote_state" "{}" {{'.format(name))
        remote_data_config.append('  backend = "s3"')
        remote_data_config.append("  config = {")
        remote_data_config.append('    region = "{}"'.format(args.state_region))
        remote_data_config.append('    bucket = "{}"'.format(args.s3_bucket))
        remote_data_config.append(
            '    key = "{}/{}/terraform.tfstate"'.format(args.s3_prefix, name)
        )
        remote_data_config.append("  }")
        remote_data_config.append("}\n")
    return "\n".join(remote_data_config)


def render_providers(providers, args):
    """Return a string that provides the provider configuration."""
    # prov_string as a list is funny sounding, but it gets joined and returned as a string
    prov_string = []
    for provider in providers:
        provider_vars = {}
        try:
            for k, v in providers[provider]["vars"].items():
                provider_vars[k] = v
        except (KeyError, TypeError):
            """No provider vars were set."""
            pass
        prov_string.append('provider "{}" {{'.format(provider))
        for k, v in provider_vars.items():
            prov_string.append('  {} = "{}"'.format(k, v))
        prov_string.append("}")
    return "\n".join(prov_string)


def run(
    name,
    temp_dir,
    terraform_path,
    command,
    key_id,
    key_secret,
    key_token=None,
    debug=False,
    plan_action="apply",
    b64_encode=False,
):
    """Run terraform."""
    params = {
        "init": "-input=false -no-color",
        "plan": "-input=false -detailed-exitcode -no-color",
        "apply": "-input=false -no-color -auto-approve",
        "destroy": "-input=false -no-color -force",
    }

    if plan_action == "destroy":
        params["plan"] += " -destroy"

    env = os.environ.copy()
    env["AWS_ACCESS_KEY_ID"] = key_id
    env["AWS_SECRET_ACCESS_KEY"] = key_secret
    if key_token is not None:
        env["AWS_SESSION_TOKEN"] = key_token
    env["TF_PLUGIN_CACHE_DIR"] = "{}/terraform-plugins".format(temp_dir)

    working_dir = "{}/definitions/{}".format(temp_dir, name)
    command_params = params.get(command)
    if not command_params:
        raise ValueError(
            "invalid command passed to terraform, {} has no defined params!".format(
                command
            )
        )

    # only execute hooks for apply/destroy
    try:
        if check_hooks("pre", working_dir, command) and command in ["apply", "destroy"]:
            # pre exec hooks
            # want to pass remotes
            # want to pass tf_vars
            click.secho(
                "found pre-{} hook script for definition {}, executing ".format(
                    command, name
                ),
                fg="yellow",
            )
            hook_exec(
                "pre",
                command,
                name,
                working_dir,
                env,
                terraform_path,
                debug=debug,
                b64_encode=b64_encode,
            )
    except HookError as e:
        click.secho(
            "hook execution error on definition {}: {}".format(name, e), fg="red"
        )
        raise SystemExit(1)

    (exit_code, stdout, stderr) = pipe_exec(
        "{} {} {}".format(terraform_path, command, command_params),
        cwd=working_dir,
        env=env,
    )
    if debug:
        click.secho("exit code: {}".format(exit_code), fg="blue")
        for line in stdout.decode().splitlines():
            click.secho("stdout: {}".format(line), fg="blue")
        for line in stderr.decode().splitlines():
            click.secho("stderr: {}".format(line), fg="red")

    # special handling of the exit codes for "plan" operations
    if command == "plan":
        if exit_code == 0:
            return True
        if exit_code == 1:
            raise TerraformError
        if exit_code == 2:
            raise PlanChange

    if exit_code:
        raise TerraformError

    # only execute hooks for plan/destroy
    try:
        if check_hooks("post", working_dir, command) and command in [
            "apply",
            "destroy",
        ]:
            # post exec hooks
            # want to pass remotes
            # want to pass tf_vars
            click.secho(
                "found post-{} hook script for definition {}, executing ".format(
                    command, name
                ),
                fg="yellow",
            )
            # hook_exec("what", "args", "are", "needed")
    except HookError as e:
        click.secho(
            "hook execution error on definition {}: {}".format(name, e), fg="red"
        )
        raise SystemExit(1)
    return True


def check_hooks(phase, working_dir, command):
    """
    check_hooks determines if a hook exists for a given operation/definition
    """
    hook_dir = "{}/hooks".format(working_dir)
    if not os.path.isdir(hook_dir):
        # there is no hooks dir
        return False
    for f in os.listdir(hook_dir):
        if os.path.splitext(f)[0] == "{}_{}".format(phase, command):
            if os.access("{}/{}".format(hook_dir, f), os.X_OK):
                return True
            else:
                raise HookError(
                    "{}/{} exists, but is not executable!".format(hook_dir, f)
                )
    return False


def hook_exec(
    phase,
    command,
    name,
    working_dir,
    env,
    terraform_path,
    debug=False,
    b64_encode=False,
):
    """
    hook_exec executes a hook script.

    Before execution it sets up the environment to make all terraform and remote
    state variables available to the hook via environment vars
    """

    key_replace_items = {
        " ": "",
        '"': "",
        "-": "_",
        ".": "_",
    }
    val_replace_items = {
        " ": "",
        '"': "",
        "\n": "",
    }
    local_env = env.copy()
    hook_dir = "{}/hooks".format(working_dir)
    hook_script = None

    for f in os.listdir(hook_dir):
        # this file format is specifically structured by the prep_def function
        if os.path.splitext(f)[0] == "{}_{}".format(phase, command):
            hook_script = "{}/{}".format(hook_dir, f)
    # this should never have been called if the hook script didn't exist...
    if hook_script is None:
        raise HookError("hook script missing from {}".format(hook_dir))

    # populate environment with terraform vars
    if os.path.isfile("{}/worker-locals.tf".format(working_dir)):
        # I'm sorry.
        r = re.compile(
            r"\s*(?P<item>\w+)\s*\=.+data\.terraform_remote_state\.(?P<state>\w+)\.outputs\.(?P<state_item>\w+)\s*"
        )

        with open("{}/worker-locals.tf".format(working_dir)) as f:
            for line in f:
                m = r.match(line)
                if m:
                    item = m.group("item")
                    state = m.group("state")
                    state_item = m.group("state_item")
                else:
                    continue

                state_value = get_state_item(
                    working_dir, env, terraform_path, state, state_item
                )

                if state_value is not None:
                    if b64_encode:
                        state_value = base64.b64encode(
                            state_value.encode("utf-8")
                        ).decode()
                    local_env[
                        "TF_REMOTE_{}_{}".format(state, item).upper()
                    ] = state_value

    # populate environment with terraform remotes
    if os.path.isfile("{}/worker.auto.tfvars".format(working_dir)):
        with open("{}/worker.auto.tfvars".format(working_dir)) as f:
            for line in f:
                tf_var = line.split("=")

                # strip bad names out for env var settings
                for k, v in key_replace_items.items():
                    tf_var[0] = tf_var[0].replace(k, v)

                for k, v in val_replace_items.items():
                    tf_var[1] = tf_var[1].replace(k, v)

                if b64_encode:
                    tf_var[1] = base64.b64encode(tf_var[1].encode("utf-8")).decode()

                local_env["TF_VAR_{}".format(tf_var[0].upper())] = tf_var[1]
    else:
        click.secho("{}/worker.auto.tfvars not found!".format(working_dir), fg="red")

    # execute the hook
    (exit_code, stdout, stderr) = pipe_exec(
        "{} {} {}".format(hook_script, phase, command), cwd=hook_dir, env=local_env,
    )

    # handle output from hook_script
    if debug:
        click.secho("exit code: {}".format(exit_code), fg="blue")
        for line in stdout.decode().splitlines():
            click.secho("stdout: {}".format(line), fg="blue")
        for line in stderr.decode().splitlines():
            click.secho("stderr: {}".format(line), fg="red")

    if exit_code != 0:
        raise HookError("hook script {}")

    pass


def get_state_item(working_dir, env, terraform_path, state, item):
    """
    get_state_item returns json encoded output from a terraform remote state
    """
    base_dir, _ = os.path.split(working_dir)
    try:
        (exit_code, stdout, stderr) = pipe_exec(
            "{} output -json -no-color {}".format(terraform_path, item),
            cwd="{}/{}".format(base_dir, state),
            env=env,
        )
    except FileNotFoundError:
        # the remote state is not setup, likely do to use of --limit
        # this is acceptable, and is the responsibility of the hook
        # to ensure it has all values needed for safe execution
        return None

    if exit_code != 0:
        raise HookError(
            "Error reading remote state item {}.{}, details: {}".format(
                state, item, stderr
            )
        )

    if stdout is None:
        raise HookError(
            "Remote state item {}.{} is empty; This is completely unexpected, failing...".format(
                state, item
            )
        )
    json_output = json.loads(stdout)
    return json.dumps(json_output, indent=None, separators=(",", ":"))



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The providers module contains functions to interact with different cloud providers.
Currently, only AWS is supported, however this provides a model to implement other
providers in the future for things like the remote state store, etc...
"""
from contextlib import closing
from tfworker.providers import StateError, validate_state_empty
import boto3

import json
import click


class aws_config(object):
    """
    aws_config provides an object to hold the required configuration needed for AWS
    provider options. This current holds extra attributes in order to expose session
    credentials to terraform.
    """

    def __init__(
        self,
        region,
        state_region,
        deployment,
        state_bucket,
        state_prefix,
        key_id=None,
        key_secret=None,
        session_token=None,
        aws_profile=None,
        role_arn=None,
    ):
        self.__key_secret = key_secret
        self.__key_id = key_id
        self.__region = region
        self.__state_region = state_region
        self.__deployment = deployment
        self.__session_token = session_token
        self.__state_bucket = state_bucket
        self.__state_prefix = state_prefix

        session_args = dict()

        if aws_profile is not None:
            session_args["profile_name"] = aws_profile

        if key_id is not None:
            session_args["aws_access_key_id"] = key_id

        if key_secret is not None:
            session_args["aws_secret_access_key"] = key_secret

        if session_token is not None:
            session_args["aws_session_token"] = session_token

        # create the base boto session
        self.__session = boto3.Session(region_name=self.__region, **session_args)

        # handle cases for assuming the role, and create a session for the state region
        if role_arn is None:
            # if a role was not provided, need to ensure credentials are set in the config, these will come from the session
            self.__key_id = self.__session.get_credentials().access_key
            self.__key_secret = self.__session.get_credentials().secret_key
            self.__session_token = self.__session.get_credentials().token

            if state_region == region:
                self.__state_session = self.__session
            else:
                self.__state_session = boto3.Session(
                    region_name=self.__state_region, **session_args
                )
        else:
            (self.__session, creds) = get_assumed_role_session(self.__session, role_arn)
            self.__key_id = creds["AccessKeyId"]
            self.__key_secret = creds["SecretAccessKey"]
            self.__session_token = creds["SessionToken"]

            if state_region == region:
                self.__state_session = self.__session
            else:
                self.__state_session = boto3.Session(
                    region_name=self.__state_region,
                    aws_access_key_id=self.__key_id,
                    aws_secret_access_key=self.__key_secret,
                    aws_session_token=self.__session_token,
                )

    @property
    def key_secret(self):
        return self.__key_secret

    @property
    def key_id(self):
        return self.__key_id

    @property
    def session_token(self):
        return self.__session_token

    @property
    def state_bucket(self):
        return self.__state_bucket

    @property
    def state_prefix(self):
        return self.__state_prefix

    @property
    def region(self):
        return self.__region

    @property
    def state_region(self):
        return self.__state_region

    @property
    def deployment(self):
        return self.__deployment

    @property
    def session(self):
        return self.__session

    @property
    def state_session(self):
        return self.__state_session


def clean_bucket_state(config, definition=None):
    """
    clean_state validates all of the terraform states are empty,
    and then removes the state objects from S3

    optionally definition can be passed to limit the cleanup
    to a single definition
    """

    s3_paginator = config.state_session.client("s3").get_paginator("list_objects_v2")
    s3_client = config.state_session.client("s3")
    if definition is None:
        prefix = config.state_prefix
    else:
        prefix = "{}/{}".format(config.state_prefix, definition)

    for s3_object in filter_keys(s3_paginator, config.state_bucket, prefix):
        state_file = s3_client.get_object(Bucket=config.state_bucket, Key=s3_object)
        body = state_file["Body"]
        with closing(state_file["Body"]):
            state = json.load(body)

        if validate_state_empty(state):
            delete_with_versions(config, s3_object)
            click.secho("state file removed: {}".format(s3_object), fg="yellow")
        else:
            raise StateError("state at: {} is not empty!".format(s3_object))


def clean_locking_state(config, deployment, definition=None):
    """
    clean_locking_state when called removes the dynamodb table
    that holds all of the state checksums and locking table
    entries
    """
    dynamo_client = config.state_session.resource("dynamodb")

    if definition is None:
        table = dynamo_client.Table("terraform-{}".format(deployment))
        table.delete()
    else:
        # delete only the entry for a single state resource
        table = dynamo_client.Table("terraform-{}".format(deployment))
        table.delete_item(
            Key={
                "LockID": "{}/{}/{}/terraform.tfstate-md5".format(
                    config.state_bucket, config.state_prefix, definition
                )
            }
        )


def filter_keys(paginator, bucket_name, prefix="/", delimiter="/", start_after=""):
    """
    filter_keys returns just they keys that are needed
    primarily from: https://stackoverflow.com/questions/30249069/listing-contents-of-a-bucket-with-boto3
    """

    prefix = prefix[1:] if prefix.startswith(delimiter) else prefix
    start_after = (start_after or prefix) if prefix.endswith(delimiter) else start_after
    try:
        for page in paginator.paginate(
            Bucket=bucket_name, Prefix=prefix, StartAfter=start_after
        ):
            for content in page.get("Contents", ()):
                yield content["Key"]
    except TypeError:
        pass


def delete_with_versions(config, key):
    """
    delete_with_versions should handle object deletions, and all references / versions of the object

    note: in initial testing this isn't required, but is inconsistent with how S3 delete markers, and the boto
    delete object call work there may be some configurations that require extra handling.
    """
    s3_client = config.state_session.client("s3")
    s3_client.delete_object(Bucket=config.state_bucket, Key=key)


def get_assumed_role_session(
    session, role_arn, session_name="AssumedRoleSession1", duration=3600
):
    """ get_assumed_role_session returns a boto3 session updated with assumed role credentials """
    sts_client = session.client("sts")
    role_creds = sts_client.assume_role(
        RoleArn=role_arn, RoleSessionName=session_name, DurationSeconds=duration
    )["Credentials"]

    new_session = boto3.Session(
        aws_access_key_id=role_creds["AccessKeyId"],
        aws_secret_access_key=role_creds["SecretAccessKey"],
        aws_session_token=role_creds["SessionToken"],
    )

    return new_session, role_creds


def unlock_state(config, definition):
    """
    sometimes terraform doesn't exit cleanly, and can leave
    a particular state locked. This function exists to provide
    an easy mechanism for removing the lock on a particular
    state. Terraform offers this directly, however it's not
    directly accessible when the worker is used to assemble
    terraform configurations.
    """



#!/usr/bin/env python
# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
import os
import struct

import click

from tfworker import terraform as tf
from tfworker.main import State, create_table, get_aws_id
from tfworker.providers.aws import aws_config, clean_bucket_state, clean_locking_state
from tfworker.providers import StateError

DEFAULT_CONFIG = "{}/worker.yaml".format(os.getcwd())
DEFAULT_REPOSITORY_PATH = "{}".format(os.getcwd())
DEFAULT_S3_BUCKET = "launchpad-terraform-states"
DEFAULT_S3_PREFIX = "terraform/state/{deployment}"
DEFAULT_AWS_REGION = "us-west-2"
DEFAULT_STATE_REGION = "us-west-2"
DEFAULT_TERRFORM = "/usr/local/bin/terraform"


def validate_deployment(ctx, deployment, name):
    """Validate the deployment is an 8 char name."""
    if len(name) > 16:
        click.secho("deployment must be less than 16 characters", fg="red")
        raise SystemExit(2)
    return name


def validate_host():
    """Ensure that the script is being run on a supported platform."""
    if struct.calcsize("P") * 8 != 64:
        click.secho("worker can only be run on 64 bit hosts, in 64 bit mode", fg="red")
        raise SystemExit(2)
    return True


@click.group()
@click.option(
    "--aws-access-key-id",
    required=False,
    envvar="AWS_ACCESS_KEY_ID",
    help="AWS Access key",
    default=None,
)
@click.option(
    "--aws-secret-access-key",
    required=False,
    envvar="AWS_SECRET_ACCESS_KEY",
    help="AWS access key secret",
    default=None,
)
@click.option(
    "--aws-session-token",
    required=False,
    envvar="AWS_SESSION_TOKEN",
    help="AWS access key token",
    default=None,
)
@click.option(
    "--aws-role-arn",
    envvar="AWS_ROLE_ARN",
    help="If provided, credentials will be used to assume this role (complete ARN)",
    default=None,
    required=False,
)
@click.option(
    "--aws-region",
    envvar="AWS_DEFAULT_REGION",
    default=DEFAULT_AWS_REGION,
    help="AWS Region to build in",
)
@click.option(
    "--aws-profile",
    required=False,
    envvar="AWS_PROFILE",
    help="The AWS/Boto3 profile to use",
    default=None,
)
@click.option(
    "--state-region",
    default=DEFAULT_STATE_REGION,
    help="AWS region where terraform state bucket exists",
)
@click.option(
    "--config-file", default=DEFAULT_CONFIG, envvar="WORKER_CONFIG_FILE", required=True
)
@click.option(
    "--repository-path",
    default=DEFAULT_REPOSITORY_PATH,
    envvar="WORKER_REPOSITORY_PATH",
    required=True,
    help="The path to the k8s-infra repository",
)
@click.pass_context
def cli(context, **kwargs):
    """CLI for the worker utility."""
    validate_host()
    config_file = kwargs["config_file"]
    try:
        context.obj = State(args=kwargs)
    except FileNotFoundError:
        click.secho(
            "configuration file {} not found".format(config_file), fg="red", err=True
        )
        raise SystemExit(1)


@cli.command()
@click.option(
    "--s3-bucket",
    default=DEFAULT_S3_BUCKET,
    help="The s3 bucket for storing terraform state",
)
@click.option(
    "--s3-prefix",
    default=DEFAULT_S3_PREFIX,
    help="The prefix in the bucket for the definitions to use",
)
@click.option("--limit", help="limit operations to a single definition", multiple=True)
@click.argument("deployment", callback=validate_deployment)
@click.pass_obj
def clean(
    obj, s3_bucket, s3_prefix, limit, deployment,
):  # noqa: E501
    """ clean up terraform state """
    if s3_prefix == DEFAULT_S3_PREFIX:
        s3_prefix = DEFAULT_S3_PREFIX.format(deployment=deployment)

    obj.clean = clean
    obj.add_arg("s3_bucket", s3_bucket)
    obj.add_arg("s3_prefix", s3_prefix)
    config = get_aws_config(obj, deployment)

    # clean just items if limit supplied, or everything if no limit
    if len(limit) > 0:
        for limit_item in limit:
            click.secho(
                "when using limit, dynamodb tables won't be completely dropped",
                fg="yellow",
            )
            try:
                # the bucket state deployment is part of the s3 prefix
                clean_bucket_state(config, definition=limit_item)
                # deployment name needs specified to determine the dynamo table
                clean_locking_state(config, deployment, definition=limit_item)
            except StateError as e:
                click.secho("error deleting state: {}".format(e), fg="red")
                raise SystemExit(1)
    else:
        try:
            clean_bucket_state(config)
        except StateError as e:
            click.secho("error deleting state: {}".format(e))
            raise SystemExit(1)
        clean_locking_state(config, deployment)


@cli.command()
@click.option(
    "--clean/--no-clean",
    default=True,
    help="clean up the temporary directory created by the worker after execution",
)
@click.option(
    "--apply/--no-apply",
    "tf_apply",
    default=False,
    help="apply the terraform configuration",
)
@click.option(
    "--force-apply/--no-force-apply",
    "force_apply",
    default=False,
    help="force apply without plan change",
)
@click.option(
    "--destroy/--no-destroy",
    default=False,
    help="destroy a deployment instead of create it",
)
@click.option(
    "--show-output/--no-show-output",
    default=False,
    help="shot output from terraform commands",
)
@click.option(
    "--s3-bucket",
    default=DEFAULT_S3_BUCKET,
    help="The s3 bucket for storing terraform state",
)
@click.option(
    "--s3-prefix",
    default=DEFAULT_S3_PREFIX,
    help="The prefix in the bucket for the definitions to use",
)
@click.option(
    "--terraform-bin",
    default=DEFAULT_TERRFORM,
    help="The complate location of the terraform binary",
)
@click.option(
    "--b64-encode-hook-values/--no--b64-encode-hook-values",
    "b64_encode",
    default=False,
    help="Terraform variables and outputs can be complex data structures, setting this open will base64 encode the values for use in hook scripts",
)
@click.option("--limit", help="limit operations to a single definition", multiple=True)
@click.argument("deployment", callback=validate_deployment)
@click.pass_obj
def terraform(
    obj,
    clean,
    tf_apply,
    force_apply,
    destroy,
    show_output,
    s3_bucket,
    s3_prefix,
    terraform_bin,
    b64_encode,
    limit,
    deployment,
):  # noqa: E501
    """Build a deployment."""
    if tf_apply and destroy:
        click.secho("can not apply and destroy at the same time", fg="red")
        raise SystemExit(1)
    plan_for = "apply"

    # If the default value is used, render the deployment name into it
    if s3_prefix == DEFAULT_S3_PREFIX:
        s3_prefix = DEFAULT_S3_PREFIX.format(deployment=deployment)

    obj.clean = clean
    obj.add_arg("s3_bucket", s3_bucket)
    obj.add_arg("s3_prefix", s3_prefix)
    obj.add_arg("terraform_bin", terraform_bin)

    # configuration for AWS interactions
    config = get_aws_config(obj, deployment)

    obj.add_arg(
        "aws_account_id",
        get_aws_id(config.key_id, config.key_secret, config.session_token),
    )

    click.secho("loading config file {}".format(obj.args.config_file), fg="green")
    obj.load_config(obj.args.config_file)

    click.secho("building deployment {}".format(deployment), fg="green")
    click.secho("using temporary Directory:{}".format(obj.temp_dir), fg="yellow")

    # common setup required for all definitions
    click.secho("downloading plugins", fg="green")
    tf.download_plugins(obj.config["terraform"]["plugins"], obj.temp_dir)
    tf.prep_modules(obj.args.repository_path, obj.temp_dir)
    create_table(
        "terraform-{}".format(deployment),
        config.state_region,
        config.key_id,
        config.key_secret,
        config.session_token,
    )

    tf_items = []

    # setup tf_items to capture the limit/order based on options
    if destroy:
        for name, body in reversed(obj.config["terraform"]["definitions"].items()):
            if limit and name not in limit:
                continue
            tf_items.append((name, body))
        plan_for = "destroy"
    else:
        for name, body in obj.config["terraform"]["definitions"].items():
            if limit and name not in limit:
                continue
            tf_items.append((name, body))

    for name, body in tf_items:
        execute = False
        # copy definition files / templates etc.
        click.secho("preparing definition: {}".format(name), fg="green")
        tf.prep_def(
            name,
            body,
            obj.config["terraform"],
            obj.temp_dir,
            obj.args.repository_path,
            deployment,
            obj.args,
        )

        # run terraform init
        try:
            tf.run(
                name,
                obj.temp_dir,
                terraform_bin,
                "init",
                config.key_id,
                config.key_secret,
                key_token=config.session_token,
                debug=show_output,
            )
        except tf.TerraformError:
            click.secho("error running terraform init", fg="red")
            raise SystemExit(1)

        click.secho("planning definition: {}".format(name), fg="green")

        # run terraform plan
        try:
            tf.run(
                name,
                obj.temp_dir,
                terraform_bin,
                "plan",
                config.key_id,
                config.key_secret,
                key_token=config.session_token,
                debug=show_output,
                plan_action=plan_for,
                b64_encode=b64_encode,
            )
        except tf.PlanChange:
            execute = True
        except tf.TerraformError:
            click.secho(
                "error planning terraform definition: {}!".format(name), fg="red"
            )
            raise SystemExit(1)

        if force_apply:
            execute = True

        if execute and tf_apply:
            if force_apply:
                click.secho("force apply for {}, applying".format(name), fg="yellow")
            else:
                click.secho("plan changes for {}, applying".format(name), fg="yellow")
        elif execute and destroy:
            click.secho("plan changes for {}, destroying".format(name), fg="yellow")
        elif not execute:
            click.secho("no plan changes for {}".format(name), fg="yellow")
            continue

        try:
            tf.run(
                name,
                obj.temp_dir,
                terraform_bin,
                plan_for,
                config.key_id,
                config.key_secret,
                key_token=config.session_token,
                debug=show_output,
                b64_encode=b64_encode,
            )
        except tf.TerraformError:
            click.secho(
                "error with terraform {} on definition {}, exiting".format(
                    plan_for, name
                ),
                fg="red",
            )
            raise SystemExit(1)
        else:
            click.secho(
                "terraform {} complete for {}".format(plan_for, name), fg="green"
            )


def get_aws_config(obj, deployment):
    """ returns an aws_config based on the paramenters sent to CLI """
    # build params for aws_config based on inputs

    config_args = dict()
    if obj.args.aws_access_key_id is not None:
        config_args["key_id"] = obj.args.aws_access_key_id

    if obj.args.aws_secret_access_key is not None:
        config_args["key_secret"] = obj.args.aws_secret_access_key

    if obj.args.aws_session_token is not None:
        config_args["session_token"] = obj.args.aws_session_token

    if obj.args.aws_profile is not None:
        config_args["aws_profile"] = obj.args.aws_profile

    if obj.args.aws_role_arn is not None:
        config_args["role_arn"] = obj.args.aws_role_arn

    config = aws_config(
        obj.args.aws_region,
        obj.args.state_region,
        deployment,
        obj.args.s3_bucket,
        obj.args.s3_prefix,
        **config_args
    )
    return config
